import unittest


class TestRHBroker(unittest.TestCase):
    def test_trade(self):
        """
        Test that it can buy and sell crypto
        """
        pass


if __name__ == "__main__":
    unittest.main()
